"""
Allow to deploy project.
Possible destinations: PYPI.

Check `deploy_to_pypi` function docs for how to use it.

Usually this function is not called manually, but it's a part of `push_pipeline` from utils.

Check utils docs where is described, how to use VS Code Task to be able to optionally test, push and deploy
with tasks (one button click).
"""
import subprocess
import os
import mylogging
import shutil

from . import misc


def deploy_to_pypi(setup_path=None):
    """Publish python library to Pypi. Username and password are set
    with env vars `TWINE_USERNAME` and `TWINE_PASSWORD`.

    Note:
        You need working `setup.py` file. If you want to see example, try the one from project-statrer on

        https://github.com/Malachov/mypythontools/blob/master/content/project-starter/setup.py

    Args:
        setup_path((str, pathlib.Path), optional): Function suppose, that there is a setup.py somewhere in cwd.
            If not, pass path to setup.py. Defaults to None.
    """

    usr = os.environ.get("TWINE_USERNAME")
    pas = os.environ.get("TWINE_PASSWORD")

    if not usr or not pas:
        raise KeyError(
            mylogging.return_str("Setup env vars TWINE_USERNAME and TWINE_PASSWORD to use deploy.")
        )

    setup_path = misc.root_path if not setup_path else setup_path

    setup_py_path = setup_path / "setup.py"

    if not setup_py_path.exists():
        setup_py_path = misc.find_path("setup.py", exclude=["node_modules"])
        setup_path = setup_py_path.parent

        if not (setup_path / "setup.py").exists():
            raise FileNotFoundError(
                mylogging.return_str("Setup.py file not found. Setup `setup_path` param.")
            )

    try:
        shutil.rmtree(setup_path / "dist")
        shutil.rmtree(setup_path / "build")

    except Exception:
        pass

    build_command = "python setup.py sdist bdist_wheel"

    subprocess.run(build_command.split(), cwd=setup_path, shell=True, check=True)

    subprocess.run(
        [
            "twine",
            "upload",
            "-u",
            os.environ["TWINE_USERNAME"],
            "-p",
            os.environ["TWINE_PASSWORD"],
            "dist/*",
        ],
        cwd=setup_path,
        shell=True,
        check=True,
    )

    shutil.rmtree(setup_path / "dist")
    shutil.rmtree(setup_path / "build")



"""
Common functions for Python / Vue / Eel project.

It contains functions for running eel, overriding eel.expose decorator, converting json to correct python format or transform data
into form for vue tables and plots.

Go on

https://mypythontools.readthedocs.io/#project-starter

for example with working examples.

Image of such an app

.. image:: /_static/project-starter-gui.png
    :width: 620
    :alt: project-starter-gui
    :align: center

"""

import os
import sys
from pathlib import Path
import warnings
import numpy as np

import mylogging

from . import misc

# Lazy imports
# import inspect

# import pandas as pd
# import EelForkExcludeFiles as eel


eel = None

expose_error_callback = None


def run_gui(devel=None, log_file_path=None, is_multiprocessing=False, builded_gui_path="default"):
    """Function that init and run `eel` project.

    It will autosetup chrome mode (if installed chrome or chromium, open separate window with
    no url bar, no bookmarks etc...) if chrome is not installed, it open microsoft Edge (by default
    on windows).

    In devel mode, app is connected on live vue server. Serve your web application with node, debug python app file
    that calls this function (do not run, just debug - server could stay running after close and occupy used port). Open browser on 8080.

    Debugger should correctly stop at breakpoints if frontend run some python function.

    Note:
        Check project-starter on github for working examples and tutorial how to run.

        https://mypythontools.readthedocs.io/#project-starter

    Args:
        devel((bool, None), optional): If None, detected. Can be overwritten. Devel 0 run static assets, 1 run Vue server on localhost.
            Defaults to None.
        log_file_path ((str, Path, None)), optional): If not exist, it will create, if exist, it will append,
            if None, log to relative log.log and only if in production mode. Defaults to None.
        is_multiprocessing (bool, optional): If using multiprocessing in some library, set up to True. Defaults to False.
        builded_gui_path ((str, Path, None)), optional): Where the web asset is. Only if debug is 0 but not run with pyinstaller.
            If None, it's automatically find (but is slower then). If 'default', path from project-starter is used - 'gui/web_builded'
            is used. Defaults to 'default'.
    """

    # Just for lazy load of eel for users that will not use this module
    global eel

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", module="EelForkExcludeFiles", category=ResourceWarning)

        import EelForkExcludeFiles as eel_library

    eel = eel_library

    try:
        if devel is None:
            # env var MY_PYTHON_VUE_ENVIRONMENT is configured and added with pyinstaller automatically in build module
            devel = False if os.environ.get("MY_PYTHON_VUE_ENVIRONMENT") == "production" else True

        # Whether run is from .exe or from python
        is_builded = True if getattr(sys, "frozen", False) else False

        if log_file_path:
            log_file = log_file_path
        else:
            log_file = "log.log" if is_builded else None

        mylogging.config.TO_FILE = log_file

        if is_builded:
            # gui folder is created with pyinstaller in build
            gui_path = Path(sys._MEIPASS) / "gui"
        else:
            if devel:
                misc.set_root()
                gui_path = (
                    misc.find_path("index.html", exclude=["node_modules", "build", "dist"]).parents[1] / "src"
                )
            else:
                if builded_gui_path:
                    gui_path = Path(builded_gui_path)

                else:
                    misc.set_root()
                    gui_path = misc.find_path(
                        "index.html", exclude=["public", "node_modules", "build"]
                    ).parent

        if not gui_path.exists():
            raise FileNotFoundError("Web files not found, setup gui_path (where builded index.html is).")

        if devel:
            directory = gui_path
            app = None
            page = {"port": 8080}
            port = 8686
            init_files = [".vue", ".js", ".html"]

            def close_callback(page, sockets):
                pass

        else:
            directory = gui_path
            close_callback = None
            app = "chrome"
            page = "index.html"
            port = 0
            init_files = [".js", ".html"]

        eel.init(directory.as_posix(), init_files, exlcude_patterns=["chunk-vendors"])

        if is_multiprocessing:
            from multiprocessing import freeze_support

            freeze_support()

        mylogging.info("Py side started")

        eel.start(
            page,
            mode=app,
            cmdline_args=["--disable-features=TranslateUI"],
            close_callback=close_callback,
            host="localhost",
            port=port,
            disable_cache=True,
        ),

    except OSError:
        eel.start(
            page,
            mode="edge",
            host="localhost",
            close_callback=close_callback,
            port=port,
            disable_cache=True,
        ),

    except Exception:
        mylogging.traceback("Py side terminated...")


def expose(callback_function):
    """Wrap eel expose with try catch block and adding exception callback function
    (for printing error to frontend usually).

    Args:
        callback_function (function): Function that will be called if exposed function fails on some error.
    """

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", module="EelForkExcludeFiles", category=ResourceWarning)

        import EelForkExcludeFiles as eel

    def inner(*args, **kargs):
        try:
            return callback_function(*args, **kargs)

        except Exception:
            mylogging.traceback(f"Unexpected error in function `{callback_function.__name__}`")
            if expose_error_callback:
                expose_error_callback()

    eel._expose(callback_function.__name__, inner)


def json_to_py(json):
    """Take json / json from JS side and eval it from strings.
    If string to string, if float to float, if object then to dict.

    When to use? - If sending object as parameter in function.

    Args:
        json (dict): Object from JS.

    Returns:
        dict: Python dictionary with correct types.
    """

    evaluated = {}
    for i, j in json.items():

        if j == "true":
            j = True
        if j == "false":
            j = False

        try:
            evaluated[i] = eval(j)
        except Exception:
            evaluated[i] = j

    return evaluated


def to_vue_plotly(data, names=None):
    """Takes data (dataframe or numpy array) and transforms it to form, that vue-plotly understand.

    Links to vue-plotly:

    https://www.npmjs.com/package/vue-plotly
    https://www.npmjs.com/package/@rleys/vue-plotly  - fork for vue 3 version

    Note:
        In js, you still need to edit the function, it's because no need to have all x axis for every column.
        Download the js function from project-starter and check for example.

    Args:
        data ((np.array, pd.DataFrame)): Plotted data.
        names (list, optional): If using array, you can define names. Defaults to None.

    Returns:
        dict: Data in form for plotting in frontend.
    """

    import pandas as pd

    if isinstance(data, (np.ndarray, np.generic)):
        data = pd.DataFrame(data, columns=names)

    data = pd.DataFrame(data)

    numeric_data = data.select_dtypes(include="number").round(decimals=3)
    numeric_data = numeric_data.where(np.isfinite(numeric_data), None)
    # numeric_data = add_none_to_gaps(numeric_data)

    return {
        "x_axis": numeric_data.index.to_list(),
        "y_axis": numeric_data.values.T.tolist(),
        "names": numeric_data.columns.values.tolist(),
    }


def to_table(df, index=False):
    """Takes data (dataframe or numpy array) and transforms it to form, that vue-plotly library understands.

    Args:
        df (pd.DataFrame): Data.
        index (bool): Whether use index as first column (or not at all).

    Returns:
        dict: DatPa in form for creating table in Vuetify v-data-table.
    """
    data = df.copy()
    data = data.round(decimals=3)

    if index:
        data.reset_index(inplace=True)

    # Numpy nan cannot be send to json - replace with None
    data = data.where(~data.isin([np.nan, np.inf, -np.inf]), None)

    headers = [{"text": i, "value": i, "sortable": True} for i in data.columns]

    return {"table": data.to_dict("records"), "headers": headers}


def help_starter_pack_vue_app():
    """
    Py, Vue and eel
    ===============

    How to develop application with these tools and how to use mypythontools functions in such an app.

    It's recomended way to download ``project-starter`` where all the files for gui are already edited and ready to use.
    There are also working examples of calling from python to js and vice versa as well as example of alerting or plotting.

    https://github.com/Malachov/mypythontools/tree/master/content

    If you want to build just what is necassaty from scratch, you can use this tutorial.
    Go in web documentation on readthedocs if reading from IDE


    Structure::
    -----------

        - myproject
            - gui
                - generated with Vue CLI
            - app.py

    app.py
    ------

    >>> from mypythontools import pyvueeel
    >>> from mypythontools.pyvueeel import expose
    ...
    >>> # Expose python functions to Js with decorator
    >>> @expose
    >>>     def load_data(settings):
    >>>         return {'Hello': 1}
    >>> if __name__ == '__main__':
    >>>     pyvueeel.run_gui()

    You can return dict - will be object in js
    You can return list - will be an array in js

    **Call js function from Py**

    >>> pyvueeel.eel.myfunction()


    gui
    ---

    Generate gui folder with Vue CLI::


        npm install -g @vue/cli
        vue create gui

    Goto folder and optionally::

        vue add vuex
        vue add vuetify
        vue add router

    main.js
    -------::

        if (process.env.NODE_ENV == 'development') {

        try {
            window.eel.set_host("ws://localhost:8686");

        } catch (error) {
            document.getElementById('app').innerHTML = 'Py side is not running. Start app.py with debugger.'
            console.error(error);
        }

        Vue.config.productionTip = true
        Vue.config.devtools = true
        } else {
        Vue.config.productionTip = false
        Vue.config.devtools = false
        }

    You can expose function to be callable from python. Import and then
    window.eel.expose(function_name, 'function_name')

    .env
    ----

    Create empty files .env.development and add `VUE_APP_EEL=http://localhost:8686/eel.js`

    Create empty .env.production and add `VUE_APP_EEL=eel.js`


    index.html
    ----------

    In public folder add to index.html::

        <script type="text/javascript" src="<%= VUE_APP_EEL %>"></script>

    vue.config.js
    -------------::

        let devtool_mode

        if (process.env.NODE_ENV === "development") {
          devtool_mode = "source-map";
        } else {
          devtool_mode = false;
        }

        module.exports = {
          outputDir: "web_builded",
          transpileDependencies: ["vuetify"],
          productionSourceMap: process.env.NODE_ENV != "production",

          configureWebpack: {
            devtool: devtool_mode,
          },
        };


    Tips, trics
    -----------

    **VS Code plugins for developing**

    - npm
    - vetur
    - Vue VSCode Snippets
    - vuetify-vscode
    """

    print(help_starter_pack_vue_app.__doc__)



"""
Module where you can configure and process paths.
"""
from pathlib import Path
import sys
import mylogging


# Root is usually current working directory, if not, use `set_paths` function.
root_path = None
app_path = None
init_path = None


def set_paths(set_root_path=None, set_init_path=None):
    """Parse python project application structure, add paths to sys.path and save to mypythontools config
    for other functions to use.

    Args:
        set_root_path ((str, pathlib.Path), optional): Path to project root where tests and docs folder are.
            If None, then cwd (current working directory) is used. Defaults to None.
        set_init_path ((str, pathlib.Path), optional): Path to project `__init__.py`. If None, then first
            found `__init__.py` is used. Defaults to None.
    """
    global init_path
    global app_path

    set_root(set_root_path=set_root_path)

    init_path = find_path("__init__.py", root_path) if not set_init_path else Path(set_init_path)
    app_path = init_path.parent


def set_root(set_root_path=None):
    """Set project root path and add it to sys.path if it's not already there.

    Args:
        set_root_path ((str, pathlib.Path), optional): Path to project root where tests and docs folder are.
            If None, then cwd (current working directory) is used. Defaults to None.
    """
    global root_path

    root_path = Path(set_root_path) if set_root_path else Path.cwd()

    if not root_path.as_posix() in sys.path:
        sys.path.insert(0, root_path.as_posix())


def find_path(file, folder=None, exclude=["node_modules", "build", "dist"], levels=5):
    """Search for file in defined folder (cwd() by default) and return it's path.

    Args:
        file (str): Name with extension e.g. "app.py".
        folder (str, optional): Where to search. If None, then root_path is used (cwd by default). Defaults to None.
        exclude (str, optional): List of folder names (anywhere in path) that will be ignored. Defaults to ['node_modules', 'build', 'dist'].
        levels (str, optional): Recursive number of analyzed folders. Defaults to 5.

    Returns:
        Path: Path of file.

    Raises:
        FileNotFoundError: If file is not found.
    """

    folder = root_path if not folder else Path(folder).resolve()

    for lev in range(levels):
        glob_file_str = f"{'*/' * lev}{file}"

        for i in folder.glob(glob_file_str):
            isthatfile = True
            for j in exclude:
                if j in i.parts:
                    isthatfile = False
                    break
            if isthatfile:
                return i

    # If not returned - not found
    raise FileNotFoundError(mylogging.return_str(f"File `{file}` not found"))


def get_desktop_path():
    """Get desktop path.

    Returns:
        Path: Return pathlib Path object. If you want string, use `.as_posix()`
    """
    return Path.home() / "Desktop"



""" Test module. Auto pytest that can be started in IDE or with

    >>> python -m pytest

in terminal in tests folder.
"""
#%%

from pathlib import Path
import os
import inspect
import shutil
import sys

import mylogging

mylogging.config.COLOR = 0

# Find paths and add to sys.path to be able to import local modules
test_path = Path(os.path.abspath(inspect.getframeinfo(inspect.currentframe()).filename)).parent
root_path = test_path.parent

if root_path not in sys.path:
    sys.path.insert(0, root_path.as_posix())

import mypythontools


def test_it():

    shutil.rmtree(root_path / "build", ignore_errors=True)
    if (root_path / "docs" / "source" / "modules.rst").exists():
        (root_path / "docs" / "source" / "modules.rst").unlink()  # missing_ok=True from python 3.8 on...

    mypythontools.misc.set_paths()
    mypythontools.utils.sphinx_docs_regenerate()
    mypythontools.utils.get_version()

    # TODO test if correct

    # Build app with pyinstaller example
    mypythontools.misc.set_paths(set_root_path=test_path)
    mypythontools.build.build_app(main_file="app.py", console=True, debug=True, cleanit=False)
    mypythontools.misc.set_paths()

    passed = (test_path / "dist").exists()

    shutil.rmtree(root_path / "tests" / "build")
    shutil.rmtree(root_path / "tests" / "dist")

    assert passed


mypythontools.misc.set_paths()
mypythontools.utils.sphinx_docs_regenerate()
